using System;
using System.Collections.Generic;
using System.Text;


namespace formula
{
	class Formula
	{


		//Zq
		private static readonly char[] fmOperator = {	'(',
														')',
														'^',
														'*',
														'/',
														'+',
														'-'};
			
		private static readonly string[] functionName={
							"ABS",	//	32
							"ACOS",	//  33
							"ASIN",	//  34
							"ATAN",	//	35
							"COS",	//	36
							"SIN",	//  37
							"TAN",	//	38
							"LN",	//	39
							"EXP",	//	40
							"LOG",	//	41
							"PI",	//	42
							"SQRT",	//	43
						};

		private const int TOKEN_ERROR		= 0;

		private const int OP_LEFT_PAREN		= 1;
		private const int OP_RIGHT_PAREN	= 2;
		private const int OP_POWER			= 3;
		private const int OP_MUL			= 4;
		private const int OP_DIV			= 5;
		private const int OP_ADD			= 6;
		private const int OP_SUB			= 7;
		private const int OP_PLUS			= 8;
		private const int OP_MINUS			= 9;

		private const int FUNC_ABS			= 32;
		private const int FUNC_ACOS			= 33;
		private const int FUNC_ASIN			= 34;
		private const int FUNC_ATAN			= 35;
		private const int FUNC_COS			= 36;
		private const int FUNC_SIN			= 37;
		private const int FUNC_TAN			= 38;
		private const int FUNC_LN			= 39;
		private const int FUNC_EXP			= 40;
		private const int FUNC_LOG			= 41;
		private const int FUNC_PI			= 42;
		private const int FUNC_SQRT			= 43;

		private const int NUMERIC_VALUE		= 1024;
		private const int STRING_VALUE		= 1025;

		public Formula()
		{
			formulaToken = new FormulaToken();
			rpnValueStack = new Stack<RpnValue>();
			//rpnValueList = new LinkedList<RpnValue>();
			rpnValueQueue = new Queue<RpnValue>();
			lastRpnValue.type = 0;
		}

		public double EvalFormula(string str)
		{
			//rpnValueList.Clear();
			rpnValueQueue.Clear();
			rpnValueStack.Clear();

			formulaToken.SetFormula(str);

			//evalPhase = 0;
			string token;
			for (; ; ) {
				int tokenType;
				token = formulaToken.GetToken(out tokenType);
				if (token.Length == 0) break;

				System.Diagnostics.Debug.WriteLine("TOKEN:" + token + " TYPE" + tokenType);

				if(tokenType == 0){
					//return false;
					return Double.NaN;
				} else if (tokenType == NUMERIC_VALUE) {
					double d;
					try {
						d = Convert.ToDouble(token);
					}
					catch {
						d = 0.0;
					}
					SetValue(d);
				} else if (tokenType >= 1 && tokenType <= 31) {
					if (!SetOperatorFunction(tokenType)) {
						//return false;
						return Double.NaN;
					}
				} else if (tokenType == STRING_VALUE) {
					int funcNum = FormulaToken.IsFunction(token);
					if (funcNum != 0) {
						SetOperatorFunction(funcNum);
					} else {
						//return false;
						return Double.NaN;
					}
				}
			}
			EndPhase1();//X^bNɎcRpnValueXgɈڂ

			return DoRpn();
		}

		//l͂ƂD揇ʂ߁AɃX^bNɐς
		/// <summary>
		/// 
		/// </summary>
		/// <param name="value"></param>
		private void SetValue(double value)
		{
			RpnValue rpnValue;
			rpnValue.priority = 70;
			rpnValue.type = NUMERIC_VALUE;
			rpnValue.value = value;
			lastRpnValue = rpnValue;		//ŌɍrpnValueo[ϐɕۑ
			rpnValueStack.Push(rpnValue);
		}
		

		private bool SetOperatorFunction(int opNum)
		{
			RpnValue rpnValue = new RpnValue();
			rpnValue.value = 0;
			rpnValue.type = opNum;

			if(opNum<32){
				switch(opNum){
				case OP_LEFT_PAREN:
					rpnValue.priority = 0;
					break;
				case OP_RIGHT_PAREN:
					rpnValue.priority = 0;
					break;
				case OP_POWER:
					rpnValue.priority = 30;
					break;
				case OP_MUL:
				case OP_DIV:
					rpnValue.priority = 20;
					break;
				case OP_ADD:
				case OP_SUB:
					if(lastRpnValue.type==NUMERIC_VALUE||					//Oo^l
					   lastRpnValue.type==OP_RIGHT_PAREN||					//Oo^
					   lastRpnValue.type>=32&&lastRpnValue.type<=1023 ){	//Oo^֐
								//PłȂ
						rpnValue.priority=10;
					}else{
								//PZq̏ꍇ
						if (lastRpnValue.type == OP_ADD) {
								rpnValue.type =OP_PLUS;
								rpnValue.priority=0;
						}else{
								rpnValue.type=OP_MINUS;
								rpnValue.priority=40;
						}
					}
					break;
				default:
					return false;
				}
			}else{
				//opNum͊֐
				rpnValue.priority=50;
			}

			lastRpnValue = rpnValue;	//ŌɍrpnValueo[ϐɕۑ

			//X^bN̑
			//Iɂ͗Dx̍̂X^bNɂX^bNoFIFOɓ
			//) 4 + 6 / 2
			//      X^bN            FIFO
			//---------------------------------
			// (1)      4                          l͂ƂDxŃX^bN
			// (2)                      4          + 萔l̗Dx̂ŃX^bNoăXg
			// (3)      +                          + X^bN
			// (4)      + 6                        l͂ƂDxŃX^bN
			// (5)      +               4 6        + 萔l̗Dx̂ŃX^bNoăXg
			// (6)      + /                        / X^bN
			// (7)      + / 2                      l͂ƂDxŃX^bN
			// (8)                      4 6 2 / +
			//
			//O͒PZA
			//PZ͉Ȃ
			//Eʂƍʂ܂ŃX^bN烊XgփRs[
			//AEʁAʎg̓XgɓȂ


			if (rpnValue.type == OP_PLUS) {					//PZȂ͉Ȃ
				return true;
			} else if (rpnValue.type == OP_LEFT_PAREN) {	//Ȃ疳ŃX^bNɐς
				rpnValueStack.Push(rpnValue);
			} else if (rpnValue.type == OP_RIGHT_PAREN) {	//EȂAX^bN獶܂POPēo^
				RpnValue popValue;
				popValue.type = 0;
				while (rpnValueStack.Count > 0) {
					popValue = rpnValueStack.Pop();
					if (popValue.type == OP_LEFT_PAREN) break;
					//rpnValueList.AddLast(popValue);
					rpnValueQueue.Enqueue(popValue);
				}
				if (popValue.type != OP_LEFT_PAREN){
					return false;	//ʂΉĂȂG[
				}
			} else {
				RpnValue popValue;
				while (rpnValueStack.Count > 0) {
					popValue = rpnValueStack.Peek();
					if (popValue.priority < rpnValue.priority) {//PeekĂ̂VlDxႢƂ
						break;									
					} else {
						popValue = rpnValueStack.Pop();
						//rpnValueList.AddLast(popValue);
						rpnValueQueue.Enqueue(popValue);
					}
				}
				rpnValueStack.Push(rpnValue);
			}
			return true;
			//evalPhase = 1;
		}

		private void EndPhase1()
		{
			while (rpnValueStack.Count > 0) {
				RpnValue popValue = rpnValueStack.Pop();
				//rpnValueList.AddLast(popValue);
				rpnValueQueue.Enqueue(popValue);
			}
		}

		private double DoRpn()
		{
			rpnValueStack.Clear();
			//) 4 + 6 / 2
			//      X^bN        FIFO        vZ
			//--------------------------------------------
			// (8)                  4 6 2 / +
			// (9)  4 6 2           / +                     Zq܂Ő擪X^bNɈڂ     
			// (10) 4               +          / 2 6  3   񍀉Zq/̂ŌvZ
			// (11) 4  3            +                       vZʂX^bNɐς
			// (12)                            + 3 4  7   񍀉Zq+̂ŌvZ
			// (13) 7                                       vZʂX^bNɐς

			while(rpnValueQueue.Count>0){
				RpnValue rpnVal = rpnValueQueue.Dequeue();
				bool opResult = true;
				switch(rpnVal.type){
					case NUMERIC_VALUE:
						rpnValueStack.Push(rpnVal);
						break;
					case OP_POWER:
						opResult = Do_POWER();
						break;
					case OP_MUL:
						opResult = Do_MUL();
						break;
					case OP_DIV:
						opResult = Do_DIV();
						break;
					case OP_ADD:
						opResult = Do_ADD();
						break;
					case OP_SUB:
						opResult = Do_SUB();
						break;
					case OP_MINUS:
						opResult = Do_MINUS();
						break;
					case FUNC_ABS:
						opResult = Do_ABS();
						break;
					case FUNC_ACOS:
						opResult = Do_ACOS();
						break;
					case FUNC_ASIN:
						opResult = Do_ASIN();
						break;
					case FUNC_ATAN:
						opResult = Do_ATAN();
						break;
					case FUNC_COS:
						opResult = Do_COS();
						break;
					case FUNC_SIN:
						opResult = Do_SIN();
						break;
					case FUNC_TAN:
						opResult = Do_TAN();
						break;
					case FUNC_LN:
						opResult = Do_LN();
						break;
					case FUNC_EXP:
						opResult = Do_EXP();
						break;
					case FUNC_LOG:
						opResult = Do_LOG();
						break;
					case FUNC_PI:
						opResult = Do_PI();
						break;
					case FUNC_SQRT:
						opResult = Do_SQRT();
						break;
					default:
						opResult = false;
						break;
				}
				if (!opResult) {
					return Double.NaN;
				}
			}

			if (rpnValueStack.Count == 1) {
				return rpnValueStack.Pop().value;
			} else {
				return Double.NaN;
			}
		}



		private bool Do_POWER()
		{
			if (rpnValueStack.Count < 2) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			RpnValue val2 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Pow(val2.value, val1.value);
			rpnValueStack.Push(resultValue);
			return true;
		}




		private bool Do_MUL()
		{
			if (rpnValueStack.Count < 2) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			RpnValue val2 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = val2.value*val1.value;
			rpnValueStack.Push(resultValue);
			return true;
		}




		private bool Do_DIV()
		{
			if (rpnValueStack.Count < 2) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			if (val1.value == 0.0) return false;
			RpnValue val2 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = val2.value / val1.value;
			rpnValueStack.Push(resultValue);
			return true;
		}




		private bool Do_ADD()
		{
			if (rpnValueStack.Count < 2) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			RpnValue val2 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = val2.value + val1.value;
			rpnValueStack.Push(resultValue);
			return true;
		}




		private bool Do_SUB()
		{
			if (rpnValueStack.Count < 2) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			RpnValue val2 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = val2.value - val1.value;
			rpnValueStack.Push(resultValue);
			return true;
		}


		private bool Do_MINUS()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = - val1.value;
			rpnValueStack.Push(resultValue);
			return true;
		}



		private bool Do_ABS()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			if (val1.value < 0) {
				resultValue.value = -val1.value;
			} else {
				resultValue.value = val1.value;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}



		private bool Do_ACOS()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue  = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Acos(val1.value);
			if(Double.IsNaN(resultValue.value)){
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}



		private bool Do_ASIN()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Asin(val1.value);
			if (Double.IsNaN(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}



		private bool Do_ATAN()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Atan(val1.value);
			if (Double.IsNaN(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}

		private bool Do_SIN()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Sin(val1.value);
			if (Double.IsNaN(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}



		private bool Do_COS()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Cos(val1.value);
			if (Double.IsNaN(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}


		private bool Do_TAN()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Tan(val1.value);
			if (Double.IsNaN(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}

		private bool Do_LN()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Log(val1.value);
			if (Double.IsNaN(resultValue.value)||
				Double.IsInfinity(resultValue.value)){
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}

		private bool Do_EXP()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Exp(val1.value);
			if (Double.IsNaN(resultValue.value) ||
				Double.IsInfinity(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}


		private bool Do_LOG()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Log10(val1.value);
			if (Double.IsNaN(resultValue.value) ||
				Double.IsInfinity(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}

		private bool Do_PI()
		{
			RpnValue resultValue = new RpnValue();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.PI;
			rpnValueStack.Push(resultValue);
			return true;
		}


		private bool Do_SQRT()
		{
			if (rpnValueStack.Count < 1) return false;
			RpnValue resultValue = new RpnValue();
			RpnValue val1 = rpnValueStack.Pop();
			resultValue.type = NUMERIC_VALUE;
			resultValue.value = Math.Sqrt(val1.value);
			if (Double.IsNaN(resultValue.value) ||
				Double.IsInfinity(resultValue.value)) {
				return false;
			}
			rpnValueStack.Push(resultValue);
			return true;
		}







		private FormulaToken formulaToken;
		Stack<RpnValue> rpnValueStack;
		//LinkedList<RpnValue> rpnValueList;
		Queue<RpnValue> rpnValueQueue;
		RpnValue lastRpnValue;
		//int evalPhase;

		private struct RpnValue
		{
			public int priority;
			public int type;
			public double value;
		}


		/// <summary>
		/// g[N؂蕪
		/// </summary>
		private class FormulaToken
		{
			public FormulaToken()
			{
				inputFormula = "";
				currentIndex = 0;
				errorAtIndex = -1;
			}

			public void SetFormula(string str)
			{
				inputFormula = str.ToUpper();
				currentIndex = 0;
				errorAtIndex = -1;
			}

			public string GetToken(out int tokenType)
			{
				tokenType = 0;
				if(inputFormula.Length==currentIndex) return "";
				StringBuilder sb = new StringBuilder();

				char c;
				//g[N擪̋󔒕̓ǂݔ΂
				do{
					if (inputFormula.Length == currentIndex) return "";
					c = inputFormula[currentIndex];
					currentIndex++;
				}while(Char.IsWhiteSpace(c));

				int n;

				//g[N̐擪ZqȂAg[N̂1Zq݂̂Ƃ݂Ȃ
				n = IsOperator(c);
				if (n > 0) {
					sb.Append(c);
					tokenType = n;
					return sb.ToString();
				}

				//g[N̐擪ȂAg[N͐lłƂ݂Ȃ
				if(Char.IsDigit(c)){
					tokenType = NUMERIC_VALUE;
					bool eflag = false;	//w\
					bool pflag = false; //_
					sb.Append(c);
					for(;;){
						//g[N̐lǂݍ
						for(;;){
							if (inputFormula.Length == currentIndex) {
								c = ' ';
								break;
							}
							c = inputFormula[currentIndex];
							if(Char.IsDigit(c)){
								sb.Append(c);
								currentIndex++;	//lł邱Ƃ킩ĂCfbNXi߂
							}else{
								break;
							}
						}
						//I[̂zCgXy[XZqȂI
						if(Char.IsWhiteSpace(c) || (IsOperator(c)!=0)){
							return sb.ToString();
						}
						if(c=='.'||c==','){ //_
							if(pflag || eflag){	//łɏ_gĂ邩w\n܂Ƃ̓G[
								errorAtIndex = currentIndex;
								tokenType = TOKEN_ERROR;
								return sb.ToString();
							}
							c = '.';
							pflag = true;
							sb.Append(c);
							currentIndex++;
							continue;
						}
						if(c=='E'){//w\
							if(eflag){	//łɎw\gĂꍇ̓G[
								errorAtIndex = currentIndex;
								tokenType = TOKEN_ERROR;
								return sb.ToString();
							}
							eflag = true;
							sb.Append(c);
							currentIndex++;
							if (inputFormula.Length == currentIndex) {
								c = ' ';
							} else {
								c = inputFormula[currentIndex];
							}
							if(c!='+' && c!='-'){
								errorAtIndex = currentIndex;
								tokenType = TOKEN_ERROR;
								return sb.ToString();
							}
							sb.Append(c);
							currentIndex++;
							continue;
						}
						errorAtIndex = currentIndex;
						tokenType = TOKEN_ERROR;
						return "";
					}
				}
				
				//g[N̐擪ȂAg[N͈ȌpƐȂ镶Ƃ݂Ȃ
				if(Char.IsLetter(c)){
					tokenType = STRING_VALUE;
					sb.Append(c);
					//g[Nǂݍ
					for(;;){
						if (inputFormula.Length == currentIndex) {
							c = ' ';
							break;
						}
						c = inputFormula[currentIndex];
						if(Char.IsLetterOrDigit(c)){
							sb.Append(c);
							currentIndex++;
						}else{
							break;
						}
					}
					return sb.ToString();
				}

				errorAtIndex = currentIndex;
				tokenType = TOKEN_ERROR;
				return "";
			}


			/// <summary>
			/// ͕ZqǂԂ
			/// </summary>
			/// <param name="c"></param>
			/// <returns>ZqȂ1ȏ̒l(OperatorNumber)Ԃ</returns>
			public static int IsOperator(char c)
			{
				int n = fmOperator.Length;
				for (int i = 0; i < n; i++) {
					if (c == fmOperator[i]) {
						return i + 1;
					}
				}
				return 0;
			}

			/// <summary>
			/// ͕񂪊֐ǂԂ
			/// </summary>
			/// <param name="str"></param>
			/// <returns>֐Ȃ32ȏ̒l(FunctionNumber)Ԃ</returns>
			public static int IsFunction(string str)
			{
				int n = functionName.Length;
				for (int i = 0; i < n; i++) {
					if (str.Equals(functionName[i])) {
						return i + 32;
					}
				}
				return 0;
			}

			public int ErrorAtIndex{
				get {return ErrorAtIndex;}
			}

			private string inputFormula;
			private int currentIndex;
			private int errorAtIndex;	//G[CfbNX
		}
														
	}
}
